//
//   @file    nnlo_utils.cxx
//
//
//   @author M.Sutton
//
//   Copyright (C) 2013 M.Sutton (sutt@cern.ch)
//
//   $Id: nnlo_utils.cxx, v0.0   Tue 13 Aug 2013 12:20:03 CEST sutt $


#include <cstdlib>
#include <cmath>
#include <string>
#include <fstream>
#include <sys/stat.h>
#include <iostream>

#include "amconfig.h"

#include "nnlo_utils.h"
#include "nnlo_common.h"


/// lhapdf
extern "C" double alphaspdf_(const double& Q);


/// sadly, includes a return at the end
std::string nnlo::date() {
  time_t _t;
  time(&_t);
  std::string a = ctime(&_t);
  std::string b = "";
  for ( unsigned i=0 ; i<a.size()-1 ; i++ ) b+=a[i];
  return b;
}



/// banner
std::string nnlo::banner() {
    std::string version = PACKAGE_VERSION;
    std::string webpage = "blafasel";//PACKAGE_URL;
    std::string banner  = "\n";
    banner += "**************************************************************************************\n";
    banner += "*   NNLOJET:     E.W.N.Glover, T.Morgan, ...                 [arXiv:.....]          *\n";
    banner += "*   nnlo-bridge: E.W.N.Glover, C. Gwenlan, T.A.Morgan, K.Rabbertz, M.R.Sutton  ...  *\n";
    banner += "*   APPLgrid:    T. Carli et al.                             [arXiv:0911.2985]       *\n";
    banner += "*   fastNLO:     D. Britzger K.Rabbertz, M. Wobisch. et al. [arXiv:...]            *\n";
    banner += "**************************************************************************************\n";
    banner += "\n";

    banner += "\n";

    banner += "nnlo-bridge version " + std::string(VERSION) + " compiled " + __DATE__ + " " + __TIME__ + "\n";
    banner += "Web page: " + webpage + "\n";
    banner += "\n";
    return banner;
}




/// simple test if a file exists
bool nnlo::file_exists(const std::string& s) {
  struct stat sb;
  if ( stat( s.c_str(), &sb)==0 ) return true; // && S_ISREG(sb.st_mode ))
  else return false;
}




/// Some kinematical definitions
double nnlo::rapidity(double E, double pz) {
    return 0.5*std::log( (E+pz)/(E-pz) );
}

double nnlo::pt( double px, double py ) {
    return std::sqrt( px*px + py*py );
}

double nnlo::mass(double E, double px, double py, double pz ) {
    return std::sqrt( E*E-px*px-py*py-pz*pz );
}

#ifdef HAVE_APPL
/// divide by bin width
void nnlo::Normalise(TH1D* h) {
  for ( int ibin=1 ; ibin<=h->GetNbinsX() ; ibin++ ) {
    double width = h->GetBinLowEdge(ibin+1) - h->GetBinLowEdge(ibin);
    h->SetBinContent( ibin, h->GetBinContent(ibin)/width );
  }
}
#endif


int nnlo::getorder( int iproc ) {

  int nloops = channel_.iloops[iproc-1]; /// loop order
  // unused:
  //  int ipars  = channel_.ipars[iproc-1];  /// number of additional particles + 1
  //  int njets  = inppar_.njets;            /// number of jets

  //    if ( dbg ) std::cout << "getorder: "
  //                     << "iproc "    << iproc
  //                     << "\tnloops " << nloops
  //                     << "\tipars "  << ipars
  //                     << "\tnjets "  << njets << " " << (nloops+ipars-njets) << std::endl;

  //  return nloops+ipars-njets; /// does not work for jets

  return ndimcurrent_.norder + nloops; /// also works for jets

}


int nnlo::GetPowLO() {
   //!< Return power of alpha_s of Born level
   if ( IsDIS() ) return inppar_.njets-1;
   if ( IsPPJ() ) return std::max(inppar_.njets,2);
   return inppar_.njets;
}


int nnlo::GetNLegs() {
   static const int& iproc = currentprocess_.iproc;
   return nnlo::GetNPow() + 1 - channel_.iloops[iproc-1];
}


int nnlo::GetOrder() {
   //!< Return order of calculation
   //!<    0:   LO
   //!<    1:  NLO
   //!<    2: NNLO

  /// this shouldn't work :
  /// currentprocess_.iproc would only ever be set
  /// once, and would then be incorrect for subsequent processes
  /// so relies on the current order *never* changing.
  /// if we really want this would be better as
  ///  static const int& iproc = ndimcurrent_.norder+channel_.iloops[ currentprocess_.iproc - 1];
  ///  return iproc;

   static const int& iproc = currentprocess_.iproc;
   return ndimcurrent_.norder+channel_.iloops[iproc-1];


   // std::cout<<"GetOrder:    "<<std::endl;
   // std::cout<<"  nord:      "<<ndimcurrent_.norder<<std::endl;
   // std::cout<<"  nv:        "<<ndimcurrent_.nv<<std::endl;
   // std::cout<<"  ipar:      "<<channel_.ipars<<std::endl;
   // //std::cout<<"  inp.njets  "<<inppar_.njets<<std::endl;

   // static const int& iproc = currentprocess_.iproc;
   // int nloops = channel_.iloops[iproc-1]; // number of loops
   // int ipars  = channel_.ipars[iproc-1];  // number of additional particles + 1
   // if ( IsDIS () )
   //    return nloops+ipars-inppar_.njets+1;
   // else
   //    return nloops+ipars-inppar_.njets;

   // if ( IsDIS() )
   //    return ndimcurrent_.norder+ndimcurrent_.nv;
   // else {
   //    if ( ndimcurrent_.nv==2 && ndimcurrent_.norder==0 ) //bugfix
   //    return ndimcurrent_.norder+ndimcurrent_.nv-1;
   //    else
   //    return ndimcurrent_.norder+ndimcurrent_.nv;
   // }
}


void nnlo::GetFillWeights( double& wgt, double& x1, double& x2, int is ) {
   //!<
   //!< Calculate the weight, and x-values that are needed by fastNLO and applgrid.
   //!< Note: call-by-reference! Do not pass the variables of the common blocks here!
   //!<

  //   using namespace std;

   // //!< Note: call-by-reference! Do not pass the variables of the common blocks here!
   // if ( &x1 == &parfrac_.x1 ) {
   //   std::cerr << "[nnlo::CalcFillWeights] BUG! Do not call CalcFillWeights with the common-block references!" << std::endl;
   //   std::exit(5);
   // }
   // if ( &x2 == &parfrac_.x2 ) {
   //   std::cerr << "[nnlo::CalcFillWeights] BUG! Do not call CalcFillWeights with the common-block references!" << std::endl;
   //   std::exit(5);
   // }


   // -------- NNLOJET, ./driver/core/sig.f rev3278 ----------- //
   //   !* remaining xreg[1|2] mapping
      // if(nv.eq.0)then ! only real integration
         // xreg1 = 0d0
         // xreg2 = 0d0// call setParFrac_kin(x1, x2)
      // else ! include virtual integration
         // xreg1 = x1 + rreg1*(1d0-x1)
         // xreg2 = x2 + rreg2*(1d0-x2)
         // wtjac = wtjac*(1d0-x1)*(1d0-x2)
         // call setParFrac_kin(x1, x2, xreg1, xreg2)
      // endif
   // -------------------------------------------------------- //

   // ---- default x from NNLOJET for tree-level
   x1 = parfrac_.x1;
   x2 = parfrac_.x2;

   //std::cout<<"x1="<<parfrac_.x1<<", x2="<<parfrac_.x2<<", xreg1="<<xregions_.xreg1<<", xreg2="<<xregions_.xreg2<<", igx="<<gridix_.igx<<", jflag="<<jacobian_.jflag<<"\twgt="<<wgt<<std::endl;

   // ---- default weights for fastNLO and APPLgrid
   wgt /= parfrac_.x1;
   if ( !nnlo::IsDIS() ) wgt /= parfrac_.x2;

   // --- fastNLO alpha_s cache
   static std::vector< std::pair < double , double > > asmur(maxscl);
   double as=0;
   if ( asmur[is].first == mur2_hook(is+1) ) as = asmur[is].second; // is>2 : mur is constant !
   else {
      as = alphaspdf_(std::sqrt(mur2_hook(is+1)));
      as = std::pow( as/(2*M_PI), GetNPow() );
      asmur[is] = std::make_pair(mur2_hook(is+1),as);
   }
   wgt /= as;// we have to divide by (alpha_s/2pi)^n

   // --- special treatment for NNLOJET
   // --- undo alpha_s multiplication
   //const double& as    = strong_.as; //or use: alphaspdf_(sqrt(mur2)); not working for scale-variations
   //const double& as    = alphaspdf_(std::sqrt(mur2_hook(is+1)));
   //wgt /= std::pow( as/(2*M_PI), GetNPow() );// we have to divide by (alpha_s/2pi)^n
   // const double& asold    = alphaspdf_(std::sqrt(mur2_hook(is+1)));
   // double wgtold = wgt/std::pow( asold/(2*M_PI), GetNPow() );// we have to divide by (alpha_s/2pi)^n


   //double wtjac   = colourflag_.colflag ? 1./xcoljac_.xcoljac :  1;
   if ( colourflag_.colflag ) wgt/= xcoljac_.xcoljac;

   // --- calculate wgt
   static const int& nv     = ndimcurrent_.nv;
   static const int& igx    = gridix_.igx;
   if ( nv==0 ){
      // only real integration
      // xreg1=0, xreg2=0
   }
   else  {
      // include virtual integration

      if ( jacobian_.jflag ) {
         wgt /= (1-parfrac_.x1);
         if ( !nnlo::IsDIS() ) wgt /= (1-parfrac_.x2);
      }

      // if ( igx==1 ) {
      // nothing to do here
      // }
      if ( igx==2 ) {
         x2  /= xregions_.xreg2;
      }
      else if ( igx==3 ) {
         x1  /= xregions_.xreg1;
      }
      else if ( igx==4 ) {
         x1 /= xregions_.xreg1;
         x2 /= xregions_.xreg2;
      }
   }

   // std::cout << "BBBBB: parfracs = " << parfrac_.x1 << ", " << parfrac_.x2 << std::endl;
   // std::cout << "BBBBB: xregions = " << xregions_.xreg1 << ", " << xregions_.xreg2 << std::endl;
   // std::cout << "BBBBB: jacobian = " << jacobian_.jflag << ", gridix = " << gridix_.igx << ", ndimcurrent nv = " << ndimcurrent_.nv << std::endl;

   // check for tiny weights
   if ( std::fabs(wgt) < TINY ) {
      wgt = 0;
      return;
   }

   if ( x1<0 || x2 <0 ) {
      std::cerr<<"ERROR. x-value is smaller than 0. x1="<<parfrac_.x1<<", x2="<<parfrac_.x2<<", xreg1="<<xregions_.xreg1<<", xreg2="<<xregions_.xreg2<<", igx="<<igx<<", jflag="<<jacobian_.jflag<<std::endl;
      wgt=0;
      x1=1;
      x2=1;
      return;
   }
   // alter weight and return;
   //wgt *= factor[order]*wtjac;

}
